import numpy as np

from pycut_pursuit.cp_d1_ql1b_cpy import cp_d1_ql1b_cpy 

def cp_d1_ql1b(Y, A, first_edge, adj_vertices, edge_weights=None, 
               Yl1=None, l1_weights=None, low_bnd=None, upp_bnd=None, 
               cp_dif_tol=1e-4, cp_it_max=10, pfdr_rho=1., pfdr_cond_min=1e-2,
               pfdr_dif_rcd=0., pfdr_dif_tol=None, pfdr_it_max=int(1e4),
               verbose=int(1e3), max_num_threads=0, max_split_size=None,
               balance_parallel_split=True, Gram_if_square=True,
               compute_List=False, compute_Graph=False, compute_Obj=False,
               compute_Time=False, compute_Dif=False):
    """
    Comp, rX, [List, Graph, Obj, Time, Dif] = cp_d1_ql1b(Y | AtY, A | AtA,
            first_edge, adj_vertices, edge_weights=None, Yl1=None,
            l1_weights=None, low_bnd=None, upp_bnd=None, cp_dif_tol=1e-4,
            cp_it_max=10, pfdr_rho=1.0, pfdr_cond_min=1e-2, pfdr_dif_rcd=0.0,
            pfdr_dif_tol=1e-2*cp_dif_tol, pfdr_it_max=int(1e4),
            verbose=int(1e3), Gram_if_square=True, max_num_threads=0,
            max_split_size=None, balance_parallel_split=True,
            compute_List=False, compute_Graph=False, compute_Obj=False,
            compute_Time=False, compute_Dif=False)

    Cut-pursuit algorithm with d1 (total variation) penalization, with a 
    quadratic functional, l1 penalization and box constraints:

    minimize functional over a graph G = (V, E)

        F(x) = 1/2 ||y - A x||^2 + ||x||_d1 + ||yl1 - x||_l1 + i_[m,M](x)

    where y in R^N, x in R^V, A in R^{N-by-|V|}
          ||x||_d1 = sum_{uv in E} w_d1_uv |x_u - x_v|,
          ||x||_l1 = sum_{v  in V} w_l1_v |x_v|,
          and the convex indicator
          i_[m,M] = infinity if it exists v in V such that x_v < m_v or 
          x_v > M_v
                  = 0 otherwise;

    using cut-pursuit approach with preconditioned forward-Douglas-Rachford 
    splitting algorithm.

    It is easy to introduce a SDP metric weighting the squared l2-norm
    between y and A x. Indeed, if M is the matrix of such a SDP metric,
    ||y - A x||_M^2 = ||Dy - D A x||^2, with D = M^(1/2).
    Thus, it is sufficient to call the method with Y <- Dy, and A <- D A.
    Moreover, when A is the identity and M is diagonal (weighted square l2 
    distance between x and y), one should call on the precomposed version 
    (see below) with Y <- DDy = My and A <- D2 = M.

    INPUTS: real numeric type is either float32 or float64, not both;
            indices numeric type can be int32 or uint32.

    NOTA: by default, components are identified using uint16 identifiers; 
    this can be easily changed in the wrapper source if more than 65535
    components are expected (recompilation is necessary)

    Y - observations, (real) array of length N (direct matricial case) or
                             array of length V (left-premult. by A^t), or
                             empty matrix (for all zeros)
    A - matrix, (real) N-by-V array (direct matricial case), or
                       V-by-V array (premultiplied to the left by A^t), or
                       V-by-1 array (_square_ diagonal of A^t A = A^2), or
                       nonzero scalar (for identity matrix), or
                       zero scalar (for no quadratic part);
        for an arbitrary scalar matrix, use identity and scale observations
        and penalizations accordingly
        if N = V in a direct matricial case, the last argument 'Gram_if_square'

      careful to the internal memory representation of multidimensional
      arrays; the C++ implementation uses column-major order (F-contiguous);
      usually numpy uses row-major order (C-contiguous), but this can often be
      taken care of without actually copying data by using transpose();

    first_edge, adj_vertices - graph forward-star representation:
        vertices are numeroted (start at 0) in the order given in Y or A
            careful to the internal memory representation of multidimensional
            arrays, usually numpy uses row-major order (C-contiguous)
        edges are numeroted (start at 0) so that all edges originating
            from a same vertex are consecutive;
        for each vertex, 'first_edge' indicates the first edge starting 
            from the vertex (or, if there are none, starting from the next 
            vertex); (int 32 or uint32) array of length V + 1, the last value
            is the total number of edges;
        for each edge, 'adj_vertices' indicates its ending vertex, (int 32 or 
            uint32) array of length E
    edge_weights - (real) array of length E or a scalar for homogeneous weights
    Yl1 - offset for l1 penalty, (real) array of length V
    l1_weights - (real) array of length V or scalar for homogeneous weights
    low_bnd - (real) array of length V or scalar
    upp_bnd - (real) array of length V or scalar
    cp_dif_tol - stopping criterion on iterate evolution; algorithm stops if
        relative changes (in Euclidean norm) is less than dif_tol;
        1e-4 is a typical value; a lower one can give better precision
        but with longer computational time and more final components
    cp_it_max - maximum number of iterations (graph cut and subproblem)
        10 cuts solve accurately most problems
    pfdr_rho - relaxation parameter, 0 < rho < 2
        1 is a conservative value; 1.5 often speeds up convergence
    pfdr_cond_min - stability of preconditioning; 0 < cond_min < 1;
        corresponds roughly the minimum ratio to the maximum descent metric;
        1e-2 is typical; a smaller value might enhance preconditioning
    pfdr_dif_rcd - reconditioning criterion on iterate evolution;
        a reconditioning is performed if relative changes of the iterate drops
        below dif_rcd; WARNING: reconditioning might temporarily draw minimizer
        away from the solution set and give bad subproblem solutions
    pfdr_dif_tol - stopping criterion on iterate evolution; algorithm stops if
        relative changes (in Euclidean norm) is less than dif_tol
        1e-2*cp_dif_tol is a conservative value
    pfdr_it_max - maximum number of iterations 1e4 iterations provides enough
        precision for most subproblems
    verbose - if nonzero, display information on the progress, every 'verbose'
        PFDR iterations
    max_num_threads - if greater than zero, set the maximum number of threads
        used for parallelization with OpenMP
    max_split_size - maximum number of vertices allowed in connected component
        passed to a split problem; make split of very large components faster,
        but might induced suboptimal artificial cuts
    balance_parallel_split - if true, the parallel workload of the split step 
        is balanced; WARNING: this might trade off speed against optimality
    Gram_if_square - if A is square, set to false for direct matricial case
    compute_List  - report the list of vertices constituting each component
    compute_Graph - get the reduced graph on the components
    compute_Obj   - compute the objective functional along iterations 
    compute_Time  - monitor elapsing time along iterations
    compute_Dif   - compute relative evolution along iterations 

    OUTPUTS: List, Graph, Obj, Time and Dif are optional, set parameters
        compute_List, compute_Graph, compute_Obj, compute_Time, or
        compute_Dif to True to request them and capture them in output
        variables in that order

    Comp - assignement of each vertex to a component, (uint16) array of
        length V
    rX - values of each component of the minimizer, (real) array of length rV;
        the actual minimizer is then reconstructed as X = rX[Comp];
    List - if requested, list of vertices constituting each component; python
        list of length rV, containing (uint32) arrays of indices
    Graph - if requested, reduced graph structure; python tuple of length 3
        representing the graph as forward-star (see input first_edge and
        adj_vertices) together with edge weights
    Obj - if requested, values of the objective functional along iterations;
        array of length actual number of cut-pursuit iterations performed + 1;
        NOTA: in the precomputed A^t A version, a constant 1/2||Y||^2 in the
        quadratic part is omited
    Time - if requested, the elapsed time along iterations; array of length
        actual number of cut-pursuit iterations performed + 1
    Dif - if requested, the iterate evolution along iterations; array of length
        actual number of cut-pursuit iterations performed

    Parallel implementation with OpenMP API.

    H. Raguet and L. Landrieu, Cut-Pursuit Algorithm for Regularizing Nonsmooth
    Functionals with Graph Total Variation, International Conference on Machine
    Learning, PMLR, 2018, 80, 4244-4253

    H. Raguet, A Note on the Forward-Douglas--Rachford Splitting for Monotone 
    Inclusion and Convex Optimization Optimization Letters, 2018, 1-24

    Baudoin Camille 2019, Hugo Raguet 2021
    """

    # Determine the type of float argument (real_t) 
    # real type is determined by Y or Yl1
    if type(Y) == np.ndarray and Y.size > 0:
        real_t = Y.dtype
    elif type(Yl1) == np.ndarray and Yl1.size > 0:
        real_t = Yl1.dtype
    else:
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: at least one of "
                        "arguments 'Y' or 'Yl1' must be provided as a nonempty"
                        " numpy array.")

    if real_t not in ["float32", "float64"]:
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: currently, the "
                        "real numeric type must be float32 or float64.")

    # Check numpy arrays: Y, A, first_edge, adj_vertices, edge_weights, Yl1,
    # l1_weights, low_bnd, upp_bnd, and define float numpy array argument with
    # the right float type if necessary
    if type(Y) != np.ndarray:
        if Y == None:
            Y = np.array([], dtype=real_t)
        else:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument 'Y' "
                            "must be a numpy array.")

    if type(A) != np.ndarray:
        if type(A) == list:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument 'A' "
                            "must be a scalar or a numpy array.")
        else:
            A = np.array([A], real_t)

    if (type(first_edge) != np.ndarray
        or first_edge.dtype not in ["int32", "uint32"]):
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
            "'first_edge' must be a numpy array of type int32 or uint32.")

    if (type(adj_vertices) != np.ndarray
        or adj_vertices.dtype not in ["int32", "uint32"]):
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
            "'adj_vertices' must be a numpy array of type int32 or uint32.")

    if type(edge_weights) != np.ndarray:
        if type(edge_weights) == list:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                    "'edge_weights' must be a scalar or a numpy array.")
        elif edge_weights != None:
            edge_weights = np.array([edge_weights], dtype=real_t)
        else:
            edge_weights = np.array([1.0], dtype=real_t)

    if type(Yl1) != np.ndarray:
        if Yl1 == None:
            Yl1 = np.array([], dtype=real_t)
        else:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                            "'Yl1' must be a numpy array.")

    if type(l1_weights) != np.ndarray:
        if type(l1_weights) == list:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                            "'l1_weights' must be a scalar or a numpy array.")
        elif l1_weights != None:
            l1_weights = np.array([l1_weights], dtype=real_t)
        else:
            l1_weights = np.array([0.0], dtype=real_t)

    if type(low_bnd) != np.ndarray:
        if type(low_bnd) == list:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                            "'low_bnd' must be a scalar or a numpy array.")
        elif low_bnd != None:
            low_bnd = np.array([low_bnd], dtype=real_t)
        else: 
            low_bnd = np.array([-np.inf], dtype=real_t)

    if type(upp_bnd) != np.ndarray:
        if type(upp_bnd) == list:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                            "'upp_bnd' must be a scalar or a numpy array.")
        elif upp_bnd != None:
            upp_bnd = np.array([upp_bnd], dtype=real_t)
        else: 
            upp_bnd = np.array([np.inf], dtype=real_t)

    # Determine V and check the graph structure
    if A.ndim > 1 and A.shape[1] != 1:
        V = A.shape[1] 
    elif A.shape[0] == 1:
        if Y.size > 0:
            V = Y.size
        elif Yl1.size > 0:
            V = Yl1.size
    else:
        V = A.shape[0]
    
    if first_edge.size != V + 1:
        raise ValueError("Cut-pursuit d1 quadratic l1 bounds: argument "
                         "'first_edge' should contain |V| + 1 = {0} elements, "
                         "but {1} are given.".format(V + 1, first_edge.size))
 
    # Check type of all numpy.array arguments of type float
    for name, ar_args in zip(
            ["Y", "A", "edge_weights", "Yl1", "l1_weights", "low_bnd",
             "upp_bnd"],
            [ Y ,  A ,  edge_weights ,  Yl1 ,  l1_weights ,  low_bnd ,
              upp_bnd ]):
        if ar_args.dtype != real_t:
            raise TypeError("argument '{0}' must be of type '{1}'"
                            .format(name, real_t))

    # Check fortran continuity of all multidimensional numpy array arguments
    if not(Y.flags["F_CONTIGUOUS"]):
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument 'Y' "
                        "must be in column-major order (F-contigous).")
    if not(A.flags["F_CONTIGUOUS"]):
        raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument 'A' "
                        "must be in column-major order (F-contigous).")

    # Convert in float64 all float arguments
    if pfdr_dif_tol is None:
        pfdr_dif_tol = 1e-2*cp_dif_tol
    cp_dif_tol = float(cp_dif_tol)
    pfdr_rho = float(pfdr_rho)
    pfdr_cond_min = float(pfdr_cond_min)
    pfdr_dif_rcd = float(pfdr_dif_rcd)
    pfdr_dif_tol = float(pfdr_dif_tol)
     
    # Convert all int arguments
    cp_it_max = int(cp_it_max)
    pfdr_it_max = int(pfdr_it_max)
    verbose = int(verbose)
    max_num_threads = int(max_num_threads)
    if max_split_size is None:
        max_split_size = V
    else:
        max_split_size = int(max_split_size)

    # Check type of all booleen arguments
    for name, b_args in zip(
            ["Gram_if_square", "balance_parallel_split", "compute_List",
             "compute_Graph", "compute_Obj", "compute_Time", "compute_Dif"],
            [ Gram_if_square ,  balance_parallel_split ,  compute_List ,
              compute_Graph ,  compute_Obj ,  compute_Time ,  compute_Dif ]):
        if type(b_args) != bool:
            raise TypeError("Cut-pursuit d1 quadratic l1 bounds: argument "
                            "'{0}' must be boolean".format(name))
    
    # Call wrapper python in C  
    return cp_d1_ql1b_cpy(Y, A, first_edge, adj_vertices, edge_weights,
        Yl1, l1_weights, low_bnd, upp_bnd, cp_dif_tol, cp_it_max, pfdr_rho,
        pfdr_cond_min, pfdr_dif_rcd, pfdr_dif_tol, pfdr_it_max, verbose,
        max_num_threads, max_split_size, balance_parallel_split,
        Gram_if_square, real_t == "float64", compute_List, compute_Graph,
        compute_Obj, compute_Time, compute_Dif)



import numpy as np

from pycut_pursuit.cp_d0_dist_cpy import cp_d0_dist_cpy

def cp_d0_dist(loss, Y, first_edge, adj_vertices, edge_weights=None, 
               vert_weights=None, coor_weights=None, cp_dif_tol=1e-3,
               cp_it_max=10, K=2, split_iter_num=2, split_damp_ratio=1.0,
               kmpp_init_num=3, kmpp_iter_num=3, min_comp_weight=0.0,
               verbose=True, max_num_threads=0, max_split_size=None,
               balance_parallel_split=True, compute_List=False,
               compute_Graph=False, compute_Obj=False, compute_Time=False,
               compute_Dif=False):
    """
    Comp, rX, [List, Graph, Obj, Time, Dif] = cp_d0_dist(loss, Y,
        first_edge, adj_vertices, edge_weights=None, vert_weights=None,
        coor_weights=None, cp_dif_tol=1e-3, cp_it_max=10, K=2,
        split_iter_num=2, split_damp_ratio=1.0, kmpp_init_num=3,
        kmpp_iter_num=3, min_comp_weight=0.0, verbose=True, max_num_threads=0,
        max_split_size=None, balance_parallel_split=True, compute_List=False,
        compute_Obj=False, compute_Time=False, compute_Dif=False)

    Cut-pursuit algorithm with d0 (weighted contour length) penalization, with
    a loss akin to a distance:

    minimize functional over a graph G = (V, E)

        F(x) = sum_v loss(y_v, x_v) + ||x||_d0

    where for each vertex, y_v and x_v are D-dimensional vectors, the loss is
    a mix of the sum of square differences and a Kullback-Leibler divergence
    (equivalent to cross-entropy in this formulation); see the 
    'loss' attribute, and ||x||_d0 = sum_{uv in E : xu != xv} w_d0_uv ,

    using greedy cut-pursuit approach with splitting initialized with 
    k-means++.

    Available data-fidelity loss include:

    quadratic (loss = D):
        f(x) = ||y - x||_{l2,W}^2 ,
    where W is a diagonal metric (separable product along ℝ^V and ℝ^D),
    that is ||y - x||_{l2,W}^2 = sum_{v in V} w_v ||x_v - y_v||_{l2,M}^2
                               = sum_{v in V} w_v sum_d m_d (x_vd - y_vd)^2;

    Kullback-Leibler divergence (equivalent to cross-entropy) on the
    probability simplex (0 < loss < 1):
        f(x) = sum_v w_v KLs(x_v, y_v),
    with KLs(y_v, x_v) = KL(s u + (1 - s) y_v ,  s u + (1 - s) x_v), where
        KL is the regular Kullback-Leibler divergence,
        u is the uniform discrete distribution over {1,...,D}, and
        s = loss is the smoothing parameter
        m is a diagonal metric weighting the coordinates;
    it yields
        KLs(y_v, x_v) = - H(s u + (1 - s) y_v)
            - sum_d (s/D + (1 - s) y_{v,d}) log(s/D + (1 - s) x_{v,d}) ,
    where H is the (weighted) entropy, that is H(s u + (1 - s) y_v)
        = - sum_d m_d (s/D + (1 - s) y_{v,d}) log(s/D + (1 - s) y_{v,d}) ;
    note that the choosen order of the arguments in the Kullback--Leibler
    does not favor the entropy of x (H(s u + (1 - s) y_v) is a constant),
    hence this loss is actually equivalent to cross-entropy.

    Both (1 <= loss < D): quadratic on coordinates from 1 to loss, and
    Kullback-Leibler divergence on coordinates from loss + 1 to D;
    
    note that coordinate weights makes no sense for Kullback-Leibler divergence
    alone, but should be used for weighting quadratic and KL when mixing both,
    in which case coor_weights should be of length loss + 1;
 
    INPUTS: real numeric type is either float32 or float64, not both;
            indices numeric type can be int32 or uint32.

    NOTA: by default, components are identified using uint16 identifiers;
    this can be easily changed in the module source if more than 65535 
    components are expected (recompilation is necessary)

    loss - D for quadratic, 0 < loss < 1 for smoothed Kullback-Leibler, D for
           quadratic, 1 <= loss < D for both (quadratic on the first 'loss'
           values and Kullback-Leibler on the 'D - loss' remaining)
    Y - observations, (real) D-by-V array;
        careful to the internal memory representation of multidimensional
        arrays; the C++ implementation uses column-major order (F-contiguous);
        usually numpy uses row-major order (C-contiguous), but this can often
        be taken care of without actually copying data by using transpose();
        for Kullback-Leibler loss, the value at each vertex must lie on the
        probability simplex
    first_edge, adj_vertices - forward-star graph representation:
        vertices are numeroted (start at 0) in the order they are given in Y;
            careful to the internal memory representation of multidimensional
            arrays, usually numpy uses row-major order (C-contiguous)
        edges are numeroted (start at 0) so that all edges originating
            from a same vertex are consecutive;
        for each vertex, 'first_edge' indicates the first edge starting from 
            the vertex (or, if there are none, starting from the next vertex);
            (int32 or uint32) array of length V+1, the last value is the
            total number of edges;
        for each edge, 'adj_vertices' indicates its ending vertex, (int32 or
            uint32) array of length E
    edge_weights - (real) array of length E or scalar for homogeneous weights
    vert_weights - weights on vertices (w_v in above notations)
        (real) array of length V or empty for no weights
    coor_weights - weights on coordinates (m_d in above notations)
        (real) array of length D (quadratic case) or of length loss + 1
        (quadratic + Kullback-Leibler case, the last coordinate weights the all
        Kullback-Leibler part), or empty for no weights
    cp_dif_tol - stopping criterion on iterate evolution; algorithm stops if
        relative changes (that is, relative dissimilarity measures defined by 
        the choosen loss between succesive iterates and between current iterate
        and observation) is less than dif_tol; 1e-3 is a typical value
    cp_it_max  - maximum number of iterations (graph cut, subproblem, merge)
        10 cuts solve accurately most problems
    K - number of alternative values considered in the split step
    split_iter_num - number of partition-and-update iterations in the split 
        step
    split_damp_ratio - edge weights damping for favoring splitting; edge
        weights increase in linear progression along partition-and-update
        iterations, from this ratio up to original value; real scalar between 0
        and 1, the latter meaning no damping
    kmpp_init_num - number of random k-means initializations in the split step
    kmpp_iter_num - number of k-means iterations in the split step
    min_comp_weight - minimum total weight (number of vertices if no weights
        are given on the vertices) that a component is allowed to have;
        components with smaller weights are merged with adjacent components
    verbose - if true, display information on the progress
    max_num_threads - if greater than zero, set the maximum number of threads
        used for parallelization with OpenMP
    max_split_size - maximum number of vertices allowed in connected component
        passed to a split problem; make split of very large components faster,
        but might induced suboptimal artificial cuts
    balance_parallel_split - if true, the parallel workload of the split step 
        is balanced; WARNING: this might trade off speed against optimality
    compute_List  - report the list of vertices constituting each component
    compute_Graph - get the reduced graph on the components
    compute_Obj   - compute the objective functional along iterations 
    compute_Time  - monitor elapsing time along iterations
    compute_Dif   - compute relative evolution along iterations 

    OUTPUTS: List, Graph, Obj, Time and Dif are optional, set parameters
        compute_List, compute_Graph, compute_Obj, compute_Time, or
        compute_Dif to True to request them and capture them in output
        variables in that order

    Comp - assignement of each vertex to a component, (uint16) array of
        length V 
    rX  - values of each component of the minimizer, (real) array of size
        D-by-rV; the actual minimizer is then reconstructed as X = rX[:, Comp];
    List - if requested, list of vertices constituting each component; python
        list of length rV, containing (uint32) arrays of indices
    Graph - if requested, reduced graph structure; python tuple of length 3
        representing the graph as forward-star (see input first_edge and
        adj_vertices) together with edge weights
    Obj - if requested, values of the objective functional along iterations;
        array of length actual number of cut-pursuit iterations performed + 1
    Time - if requested, the elapsed time along iterations; array of length
        actual number of cut-pursuit iterations performed + 1
    Dif - if requested, if requested, the iterate evolution along iterations;
        array of length actual number of cut-pursuit iterations performed
 
    Parallel implementation with OpenMP API.

    L. Landrieu and G. Obozinski, Cut Pursuit: fast algorithms to learn
    piecewise constant functions on general weighted graphs, SIAM Journal on
    Imaging Science, 10(4):1724-1766, 2017

    L. Landrieu et al., A structured regularization framework for spatially
    smoothing semantic labelings of 3D point clouds, ISPRS Journal of
    Photogrammetry and Remote Sensing, 132:102-118, 2017

    Baudoin Camille 2019, Raguet Hugo 2021, 2022
    """
    
    # Determine the type of float argument (real_t) 
    # real_t type is determined by the first parameter Y 
    if type(Y) != np.ndarray:
        raise TypeError("Cut-pursuit d0 distance: argument 'Y' must be a "
                        "numpy array.")

    if Y.size > 0 and Y.dtype == "float64":
        real_t = "float64" 
    elif Y.size > 0 and Y.dtype == "float32":
        real_t = "float32" 
    else:
        raise TypeError("Cut-pursuit d0 distance: argument 'Y' must be a "
                        "nonempty numpy array of type float32 or float64.") 
    
    # Convert in numpy array scalar entry: Y, first_edge, adj_vertices, 
    # edge_weights, vert_weights, coor_weights and define float numpy array
    # argument with the right float type, if empty:
    if (type(first_edge) != np.ndarray
        or first_edge.dtype not in ["int32", "uint32"]):
        raise TypeError("Cut-pursuit d0 distance: argument 'first_edge' must "
                        "be a numpy array of type int32 or uint32.")

    if (type(adj_vertices) != np.ndarray
        or adj_vertices.dtype not in ["int32", "uint32"]):
        raise TypeError("Cut-pursuit d0 distance: argument 'adj_vertices' "
                        "must be a numpy array of type int32 or uint32.")

    if type(edge_weights) != np.ndarray:
        if type(edge_weights) == list:
            raise TypeError("Cut-pursuit d0 distance: argument 'edge_weights' "
                            "must be a scalar or a numpy array.")
        elif edge_weights != None:
            edge_weights = np.array([edge_weights], dtype=real_t)
        else:
            edge_weights = np.array([1.0], dtype=real_t)
        
    if type(vert_weights) != np.ndarray:
        if type(vert_weights) == list:
            raise TypeError("Cut-pursuit d0 distance: argument 'vert_weights' "
                            "must be a numpy array.")
        else:
            vert_weights = np.array([], dtype=real_t)

    if type(coor_weights) != np.ndarray:
        if type(coor_weights) == list:
            raise TypeError("Cut-pursuit d0 distance: argument 'coor_weights' "
                            "must be a numpy array.")
        else:
            coor_weights = np.array([], dtype=real_t)

    # Determine V and check graph structure 
    if Y.ndim > 1:
        V = Y.shape[1]
        D = Y.shape[0]
    else:
        V = Y.shape[0]
        D = 1
        
    if first_edge.size != V + 1 :
        raise ValueError("Cut-pursuit d0 distance: argument 'first_edge'"
                         "should contain |V| + 1 = {0} elements, "
                         "but {1} are given".format(V + 1, first_edge.size))
 
    # Check type of all numpy.array arguments of type float
    for name, ar_args in zip(
            ["Y", "edge_weights", "vert_weights", "coor_weights"],
            [ Y ,  edge_weights ,  vert_weights ,  coor_weights ]):
        if ar_args.dtype != real_t:
            raise TypeError("Cut-pursuit d0 distance: argument '{0}' must be "
                            "of type '{1}'".format(name, real_t))

    # Check fortran continuity of all multidimensional numpy array arguments
    if not(Y.flags["F_CONTIGUOUS"]):
        raise TypeError("Cut-pursuit d0 distance: argument 'Y' must be in "
                        "column-major order (F-contigous).")

    # Convert in float64 all float arguments
    loss = float(loss)
    split_damp_ratio = float(split_damp_ratio)
    if 0.0 < loss and loss < 1.0 and coor_weights.size > 0:
        raise ValueError("Cut-pursuit d0 distance: with Kullback-Leibler loss,"
                     " 'coor_weights' should be empty; weighting coordinates "
                     "of the probability space makes no sense.")
    elif (1 <= loss and loss < D and coor_weights.size > 0 and
        coor_weights.size != loss + 1):
        raise ValueError("Cut-pursuit d0 distance: for weighting quadratic and"
            " Kullback-Leibler parts, 'coor_weights' should be of size loss + "
            "1, where loss indicates the number of coordinates involved in the"
            " quadratic part and the last weights is for KL")
    elif loss == D and coor_weights.size > 0 and coor_weights.size != D:
        raise ValueError("Cut-pursuit d0 distance: with quadratic loss,"
            " argument 'coor_weights' should be empty or of size D")
    cp_dif_tol = float(cp_dif_tol)
     
    # Convert all int arguments: 
    cp_it_max = int(cp_it_max)
    K = int(K)
    split_iter_num = int(split_iter_num)
    kmpp_init_num = int(kmpp_init_num)
    kmpp_iter_num = int(kmpp_iter_num)
    max_num_threads = int(max_num_threads)
    if max_split_size is None:
        max_split_size = V
    else:
        max_split_size = int(max_split_size)

    # Check type of all booleen arguments
    for name, b_args in zip(
        ["verbose", "balance_parallel_split", "compute_List", "compute_Graph",
         "compute_Obj", "compute_Time", "compute_Dif"],
        [ verbose ,  balance_parallel_split ,  compute_List ,  compute_Graph ,
          compute_Obj ,  compute_Time ,  compute_Dif ]):
        if type(b_args) != bool:
            raise TypeError("Cut-pursuit d0 distance: argument '{0}' must be "
                            "boolean".format(name))

    # Call wrapper python in C  
    return cp_d0_dist_cpy(loss, Y, first_edge, adj_vertices, edge_weights,
        vert_weights, coor_weights, cp_dif_tol, cp_it_max, K, split_iter_num,
        split_damp_ratio, kmpp_init_num, kmpp_iter_num, min_comp_weight,
        verbose, max_num_threads, max_split_size, balance_parallel_split,
        real_t == "float64", compute_List, compute_Graph, compute_Obj,
        compute_Time, compute_Dif)



# ----------------------------------------------------------------------#
#  setuptools setup script for compiling cut-pursuit python extensions  #
# ----------------------------------------------------------------------#
""" 
Compilation command: python setup.py build_ext

Camille Baudoin 2019
"""

from setuptools import setup, Extension
import numpy
import os


# Include directories
include_dirs = [
    numpy.get_include(),  # find the Numpy headers
    "pcd-prox-split/include",
    "pcd-prox-split/matrix-tools/include",
    "pcd-prox-split/proj-simplex/include",
    "wth-element/include",
    "include",
]

# Compilation and linkage options
# _GLIBCXX_PARALLEL is only useful for libstdc++ users
# MIN_OPS_PER_THREAD roughly controls parallelization, see doc in README.md
if os.name == "nt":  # windows
    extra_compile_args = ["/openmp", "-DMIN_OPS_PER_THREAD=10000"]
    extra_link_args = []
elif os.name == "posix":  # linux
    extra_compile_args = [
        "-std=c++11",
        "-fopenmp",
        "-D_GLIBCXX_PARALLEL",
        "-DMIN_OPS_PER_THREAD=10000",
    ]
    extra_link_args = ["-lgomp"]
else:
    raise NotImplementedError("OS not yet supported.")

COMP_T_ON_32_BITS = os.environ.get("COMP_T_ON_32_BITS", None)

if COMP_T_ON_32_BITS is not None and COMP_T_ON_32_BITS == "1":
    extra_compile_args.append("-DCOMP_T_ON_32_BITS")

# Compilation
mod_cp_d1_ql1b = Extension(
    "pycut_pursuit.cp_d1_ql1b_cpy",
    # list source files
    [
        "python/cpython/cp_d1_ql1b_cpy.cpp",
        "src/cp_d1_ql1b.cpp",
        "src/cut_pursuit_d1.cpp",
        "src/cut_pursuit.cpp",
        "src/maxflow.cpp",
        "pcd-prox-split/src/pfdr_d1_ql1b.cpp",
        "pcd-prox-split/src/pfdr_graph_d1.cpp",
        "pcd-prox-split/src/pcd_fwd_doug_rach.cpp",
        "pcd-prox-split/src/pcd_prox_split.cpp",
        "pcd-prox-split/matrix-tools/src/matrix_tools.cpp",
    ],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

mod_cp_d1_lsx = Extension(
    "pycut_pursuit.cp_d1_lsx_cpy",
    # list source files
    [
        "python/cpython/cp_d1_lsx_cpy.cpp",
        "src/cp_d1_lsx.cpp",
        "src/cut_pursuit_d1.cpp",
        "src/cut_pursuit.cpp",
        "src/maxflow.cpp",
        "pcd-prox-split/src/pfdr_d1_lsx.cpp",
        "pcd-prox-split/src/pfdr_graph_d1.cpp",
        "pcd-prox-split/src/pcd_fwd_doug_rach.cpp",
        "pcd-prox-split/src/pcd_prox_split.cpp",
        "pcd-prox-split/proj-simplex/src/proj_simplex.cpp",
    ],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

mod_cp_d0_dist = Extension(
    "pycut_pursuit.cp_d0_dist_cpy",
    # list source files
    [
        "python/cpython/cp_d0_dist_cpy.cpp",
        "src/cp_d0_dist.cpp",
        "src/cut_pursuit_d0.cpp",
        "src/cut_pursuit.cpp",
        "src/maxflow.cpp",
    ],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

mod_cp_prox_tv = Extension(
    "pycut_pursuit.cp_prox_tv_cpy",
    # list source files
    [
        "python/cpython/cp_prox_tv_cpy.cpp",
        "src/cp_prox_tv.cpp",
        "src/cut_pursuit_d1.cpp",
        "src/cut_pursuit.cpp",
        "src/maxflow.cpp",
        "pcd-prox-split/src/pfdr_prox_tv.cpp",
        "pcd-prox-split/src/pfdr_graph_d1.cpp",
        "pcd-prox-split/src/pcd_fwd_doug_rach.cpp",
        "pcd-prox-split/src/pcd_prox_split.cpp",
    ],
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

setup(
    package_dir={"pycut_pursuit": "python/wrappers"},
    ext_modules=[mod_cp_d1_ql1b, mod_cp_d1_lsx, mod_cp_d0_dist, mod_cp_prox_tv]
)
